# automatically generated by the FlatBuffers compiler, do not modify

# namespace: tosa

class QuantInfo(object):
    NONE = 0
    UnaryQuantInfo = 1
    ConvQuantInfo = 2
    MatMulQuantInfo = 3
    PadQuantInfo = 4

