# automatically generated by the FlatBuffers compiler, do not modify

# namespace: tosa

class Attribute(object):
    NONE = 0
    PoolAttribute = 1
    ConvAttribute = 2
    TransposeConvAttribute = 3
    ReluNAttribute = 4
    AxisAttribute = 5
    ReshapeAttribute = 6
    SliceAttribute = 7
    TileAttribute = 8
    ResizeAttribute = 9
    ClampAttribute = 10
    RescaleAttribute = 11
    MulAttribute = 12
    ArithmeticRightShiftAttribute = 13
    CondIfAttribute = 14
    WhileLoopAttribute = 15

