# automatically generated by the FlatBuffers compiler, do not modify

# namespace: tosa

import flatbuffers

class TileAttribute(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsTileAttribute(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = TileAttribute()
        x.Init(buf, n + offset)
        return x

    # TileAttribute
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # TileAttribute
    def Multiples(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Int32Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # TileAttribute
    def MultiplesAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Int32Flags, o)
        return 0

    # TileAttribute
    def MultiplesLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

def TileAttributeStart(builder): builder.StartObject(1)
def TileAttributeAddMultiples(builder, multiples): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(multiples), 0)
def TileAttributeStartMultiplesVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def TileAttributeEnd(builder): return builder.EndObject()
